import numpy as np
import matplotlib.pyplot as plt
import sys
import jax
import jax.numpy as jnp
import jax.experimental.sparse as jsp
from sparsejac import jacfwd
from scipy.linalg import norm
import scipy.sparse as ssp
jax.config.update("jax_enable_x64", True)

# for debug
import pdb
import signal

def signal_handler(sig, frame):
    choice = input("\nto debug enter 'dbg', to exit press Enter: >")
    if choice == 'dbg':
        pdb.set_trace()
    else:
        sys.exit(0)
signal.signal(signal.SIGINT, signal_handler)

"""
    Two-phase flow model from doi.org/10.1515/rnam-2023-0025 
    Cylindrical coordinates for the single setup of numerical experiments.
    Residual is assembled using jax and sparsejac (see requirements.txt)
"""


prm = {
    'L': 0.01, # Sample thickness, m
    'D': 0.025, # Sample diameter, m
    'phi0': 0.16, # Reference porosity φ0
    'Pnw0': 8e6, # Initial water pressure Pnw,0 , Pa
    'Pw0': 8e6, # Reference CO2 pressure
    'muw': 1e-5, # CO2 dynamic viscosity μw , Pa · s
    'munw': 1e-3, # Water dynamic viscosity μnw , Pa · s
    'Pt': 15e6, # Confining pressure P t , Pa
    'Patm': 0.1e6, # Atmospheric pressure, Pa
    'T': 7200, # Modelling time, s
    'Q': 2.6e-7, # Injection rate, kg/s
    'Pf': 8e6, # Downstream pressure Pf , Pa,
    'rhow0': 3e2, # Reference CO2 density
    'rhonw': 1e3, # Constant water density

    # bounds from article
    'vg_beta': (1e-5, 1), # Van Genuchten parameter \beta, Pa-1
    'vg_n': (2, 10), # Van Genuchten parameter n
    'gamma': (3.2e-9, 3.2e-7), # Exponential factor, Pa-1
    'k0': (1e-19, 1e-17), # Intrinsic permeability, m2
    'cphi': (1.2e-9, 1.2e-7), # Shale compressibility, Pa-1
    'rinj': (0.0005, 0.01), # Effective injection radius, m
    'nu': (2, 8), # Power factor for relative permeability
    'Sw0': (1e-4, 1e-1), # Initial CO2 saturation
    'cw': (1e-7, 1e-4), # CO2 compressibility, Pa-1

    'Nr': 16, # amount of cells in R
    'Nz': 16, # and Z directions
    'dt0': 0.1, # initial time step
    'dtmax': 50.0, # maximum time step
    'switch_threshold': 0.9 # if Sw > threshold, use Pw as primary variable, otherwise use Sw
}


thr = prm['switch_threshold']
L, D = prm['L'], prm['D']
rhow0, rhonw, Pw0 = prm['rhow0'], prm['rhonw'], prm['Pw0']
#! see initial condition
Pnw0 =  prm['Pf']
muw, munw = prm['muw'], prm['munw']
phi0 = prm['phi0']
Pt, Patm, Pf = prm['Pt'], prm['Patm'], prm['Pf']
Q, T = prm['Q'], prm['T']

vg_beta, vg_n = 9.492192578125e-01, 4.15625
vg_m = 1.-1./vg_n
vg_beta = vg_beta/(rhow0*9.81)
Sw0 = 5.9e-3
cphi = 1e-9
k0 = 6.9e-19
nu = 7.6
gamma = 3.1e-7
rinj = 2.8e-3
cw = 1.7e-5


Nr, Nz = prm['Nr'], prm['Nz']
hr, hz = D/2 / Nr, L / Nz
N = Nr * Nz # total amount of cells


def _pw(sw,pnw):
    return pnw - Pc(sw)
def _sw(pw,pnw):
    return invPc(pnw-pw)
def _pw1(pnw, xw, swi):
    return jnp.where(swi>0, xw, _pw(xw, pnw))
def _sw1(pnw, xw, swi):
    return jnp.where(swi>0, _sw(xw, pnw), xw)

def _rhow(pw):
    return rhow0 + cw*(pw - Pw0)
def Pc(sw):
    return 1./vg_beta * jnp.power((jnp.power(sw,-1./vg_m) - 1), 1./vg_n)
def invPc(pc):
    return jnp.power(jnp.power(vg_beta*pc, vg_n)+1, -vg_m)
def _pf(pw,pnw,sw):
    return sw*pw+(1-sw)*pnw
def pw2(sw,pf):
    return pf-(1-sw)*Pc(sw)
def pw3(pnw,sw,pf):
    return (pf-(1-sw)*pnw)/sw
def pnw2(sw,pf):
    return pf+sw*Pc(sw)
def phi1(phi0,pf,pf0):
    # porosity grows unbounded, which is unphysical
    #return phi0/(1.-cphi*(pf-pf0))
    # dPhi/dt = c_phi Phi(1-Phi) dPf/dt 
    return phi0/(1.-cphi*(1-phi0)*(pf-pf0))
def krw(sw):
    return jnp.power(sw, nu)
def krnw(sw):
    return jnp.power(1.-sw, nu)
def _ke(pf):
    return jnp.exp(-gamma*(Pt-pf-Patm))
def _phi2(pw, pnw, sw, phi0, pf0):
    return phi1(phi0, _pf(pw,pnw,sw), pf0)

def _lmbdnw1(pnw,xw,swi):
    return rhonw * krnw(_sw1(pnw,xw,swi)) * _ke(_pf1(pnw,xw,swi)) / munw
def _lmbdw1(pnw,xw,swi):
    return _rhow1(pnw,xw,swi) * krw(_sw1(pnw,xw,swi)) * _ke(_pf1(pnw,xw,swi)) / muw
def _pf1(pnw,xw,swi):
    return _pf(_pw1(pnw, xw, swi),pnw,_sw1(pnw, xw, swi))
def _krw(pnw, xw, swi):
    return krw(_sw1(pnw, xw, swi))
def _krnw(pnw, xw, swi):
    return krnw(_sw1(pnw, xw, swi))
def _ke1(pnw, xw, swi):
    return _ke(_pf1(pnw,xw,swi))
def _phi1(pnw, xw, swi, phi0, pf0):
    return phi1(phi0, _pf1(pnw,xw,swi), pf0)
def _rhow1(pnw, xw, swi):
    return _rhow(_pw1(pnw, xw, swi))
def _pw2(pnw,xw,swi):
    return pw2(_sw1(pnw,xw,swi),Pf)
def _pnw2(pnw,xw,swi):
    return pnw2(_sw1(pnw,xw,swi),Pf)
def _sw2(pnw,xw,swi):
    return sw2(pnw,_sw1(pnw,xw,swi),Pf)


@jax.jit
def residual(X, X0, ST, PHI, dt):
    rad = hr*np.arange(Nr+1)
    perim = 2*np.pi * rad[1:-1]
    sqarea = np.pi * (rad[1:]**2 - rad[:-1]**2)
    K = hz*np.tile(sqarea, Nz)

    F = jnp.zeros_like(X0)

    Phi0 = PHI[N:]
    Pnw, Xw, Swi = X[N:], X[:N], ST[:N]
    Pnw0, Xw0, Swi0 = X0[N:], X0[:N], ST[N:]
    Pw, Pw0 = _pw1(Pnw,Xw,Swi), _pw1(Pnw0,Xw0,Swi0)
    Sw, Sw0 = _sw1(Pnw,Xw,Swi), _sw1(Pnw0,Xw0,Swi0)
    Phi1 = _phi2(Pw, Pnw, Sw, phi0=Phi0, pf0=_pf(Pw0,Pnw0,Sw0))
    Rhow1, Rhow0 = _rhow(Pw), _rhow(Pw0)
    # first half (wetting) of residual
    F = F.at[:N].set(K * ( Rhow1*Sw*Phi1 - Rhow0*Sw0*Phi0 - Rhow1 * Sw*Phi1/(1-Phi1) * (Phi1-Phi0) ))
    # second half (non-wetting) of residual
    F = F.at[N:].set(K * rhonw * ( (1-Sw)*Phi1 - (1-Sw0)*Phi0 - (1-Sw)*Phi1/(1-Phi1) * (Phi1-Phi0) ))

    # loop over inner vertical faces
    transm = k0*perim*hz/hr

    for j in range(Nz):
        i1, i2 = j*Nr, (j+1)*Nr
        Pnw, Xw, Swi = X[N+i1:N+i2], X[i1:i2], ST[i1:i2]
        Pw = _pw1(Pnw,Xw,Swi)
        Lmbdnw = jnp.where(Pnw[:-1]>Pnw[1:], _lmbdnw1(Pnw[:-1],Xw[:-1],Swi[:-1]), _lmbdnw1(Pnw[1:],Xw[1:],Swi[1:]))
        Lmbdw = jnp.where(Pw[:-1]>Pw[1:], _lmbdw1(Pnw[:-1],Xw[:-1],Swi[:-1]), _lmbdw1(Pnw[1:],Xw[1:],Swi[1:]))
        F = F.at[i1:i2-1].add( dt*Lmbdw*transm * (Pw[:-1] - Pw[1:]))
        F = F.at[i1+1:i2].add( -dt*Lmbdw*transm * (Pw[:-1] - Pw[1:]))
        F = F.at[N+i1:N+i2-1].add( dt*Lmbdnw*transm * (Pnw[:-1] - Pnw[1:]))
        F = F.at[N+i1+1:N+i2].add( -dt*Lmbdnw*transm * (Pnw[:-1] - Pnw[1:]))

    # loop over inner horizontal faces
    transm = k0/hz

    for k in range(Nr):
        Pnw, Xw, Swi = X[N+k::Nr], X[k:N:Nr], ST[k:N:Nr]
        Pw = _pw1(Pnw,Xw,Swi)
        Lmbdnw = jnp.where(Pnw[:-1]>Pnw[1:], _lmbdnw1(Pnw[:-1],Xw[:-1],Swi[:-1]), _lmbdnw1(Pnw[1:],Xw[1:],Swi[1:]))
        Lmbdw = jnp.where(Pw[:-1]>Pw[1:], _lmbdw1(Pnw[:-1],Xw[:-1],Swi[:-1]), _lmbdw1(Pnw[1:],Xw[1:],Swi[1:]))
        F = F.at[k:N-Nr:Nr].add( dt*Lmbdw*transm*sqarea[k] * (Pw[:-1] - Pw[1:]))
        F = F.at[k+Nr:N:Nr].add( -dt*Lmbdw*transm*sqarea[k] * (Pw[:-1] - Pw[1:]))
        F = F.at[N+k:-Nr:Nr].add( dt*Lmbdnw*transm*sqarea[k] * (Pnw[:-1] - Pnw[1:]))
        F = F.at[N+k+Nr::Nr].add( -dt*Lmbdnw*transm*sqarea[k] * (Pnw[:-1] - Pnw[1:]))

    ind = np.arange(np.ceil(rinj/hr), dtype=int)

    # bottom boundary: pf = Pf
    j = Nz-1
    i1, i2 = j*Nr, (j+1)*Nr
    Pnw, Xw, Swi = X[N+i1+ind], X[i1+ind], ST[i1+ind]
    # Pnw, Xw, Swi = X[N+i1:N+i2], X[i1:i2], ST[i1:i2]
    Pw = _pw1(Pnw,Xw,Swi)
    # Pnw_bnd, Pw_bnd = _pnw2(Pnw, Xw, Swi), _pw2(Pnw,Xw,Swi)
    Pnw_bnd = _pnw2(Pnw, Xw, Swi)
    Pw_bnd = pw3(Pnw_bnd,_sw1(Pnw,Xw,Swi),Pf)
    Lmbdnw = jnp.where(Pnw>Pnw_bnd, _lmbdnw1(Pnw,Xw,Swi), _lmbdnw1(Pnw_bnd,Xw,Swi))
    Lmbdw = jnp.where(Pw>Pw_bnd, _lmbdw1(Pnw,Xw,Swi), _lmbdw1(Pnw_bnd,Xw,Swi))
    F = F.at[i1+ind].add( dt*Lmbdw*transm*sqarea[ind] * (Pw - Pw_bnd))
    F = F.at[N+i1+ind].add( dt*Lmbdnw*transm*sqarea[ind] * (Pnw - Pnw_bnd))
    # F = F.at[i1:i2].add( dt*Lmbdw*transm*sqarea * (Pw - Pw_bnd))
    # F = F.at[N+i1:N+i2].add( dt*Lmbdnw*transm*sqarea * (Pnw - Pnw_bnd))

    # top boundary: fixed influx -uw * n = Q
    rel = sqarea[ind]
    rel /= np.sum(rel)
    F = F.at[ind].add(-dt*Q*rel)

    return F

def neighbours(ind):
    row, col = divmod(ind, Nr)
    nbr = [row*Nr+col]
    if row > 0:
        nbr.append((row-1)*Nr + col)
    if row < Nz-1:
        nbr.append((row+1)*Nr + col)
    if col > 0:
        nbr.append(row*Nr + col-1)
    if col < Nr-1:
        nbr.append(row*Nr + col+1)
    return nbr

def num_neighbours():
    total = 0
    for ind in range(N):
        row, col = divmod(ind, Nr)
        total += 1 + (row > 0) + (row < Nz-1) + (col > 0) + (col < Nr-1)
    return total

def incidence_matrix():
    n = num_neighbours()
    indices = np.empty(shape=(4*n, 2), dtype=np.int32)
    data = np.ones(indices.shape[0], dtype=np.float64)
    for ind in range(2*N):
        indices[ind] = (ind,ind)
    k = 2*N
    for ind in range(N):
        for q in neighbours(ind % N):
            if ind != q:
                indices[k] = (ind, q)
                indices[k+1] = (ind+N, q+N)
                k += 2
            indices[k] = (ind, q+N)
            indices[k+1] = (ind+N, q)
            k += 2
    if k != indices.shape[0]:
        pdb.set_trace()
    return jsp.BCOO((data,indices), shape=(2*N,2*N))


def remove_ticks(ax):
    ax.axes.get_xaxis().set_ticks([])
    ax.axes.get_yaxis().set_ticks([])

def update_state(X,ST0):
    ST1 = ST0.copy()
    wasPw, = np.asarray(ST0>0).nonzero()
    wasSw, = np.asarray(ST0==0).nonzero()
    toPw, = np.asarray(X[wasSw] > thr).nonzero()
    toSw, = np.asarray(_sw(X[wasPw],X[N+wasPw]) < thr).nonzero()
    toPw, toSw = wasSw[toPw], wasPw[toSw]
    # if toPw.size > 0 or toSw.size > 0:
    #     pdb.set_trace()
    X = X.at[toPw].set(_pw(X[toPw],X[N+toPw]))
    X = X.at[toSw].set(_sw(X[toSw],X[N+toSw]))
    ST1[toPw] = 1
    ST1[toSw] = 0
    return X, ST1

if __name__ == "__main__":
    t, dt0, dtmax = 0.0, prm['dt0'], prm['dtmax']
    dt = dt0
    U = jnp.zeros(2*N) # Sw/Pw, Pnw
    # initial conditions: constant Sw0 and constant Pnw such that Pf0=Pf (no Pf gradient on bottom boundary)
    U = U.at[:N].set(Sw0)
    U = U.at[N:].set(pnw2(Sw0, Pf))
    #if np.abs(pnw2(Sw0, Pf)-pw2(Sw0,Pf) - Pc(Sw0)) > 1.0e-6:
    #    pdb.set_trace()
    U0 = np.array(U)
    ST = np.zeros(2*N, dtype=np.int8) # state, state0 of wetting fluid in cell (0 for Sw if Sw < 0.9 else 1 for Pw)
    ST2 = np.zeros(N, dtype=np.int8) # temporary state vector
    PHI = phi0*np.ones(2*N) # phi, phi0

    fig = None

    # time loop
    rst, maxrst = 0, 7 # amount of restarts
    spars = incidence_matrix()
    while t < T + 1.0e-10:
        fail, converged, restart = False, False, False
        it, maxit = 1, 50 # Newton iterations, maximum amount of iterations allowed before restart
        lrtol, latol = 1e-8, 1e-12 # linear solver relative and absolute tolerances
        nrtol, natol = 1e-5, 1e-9 # Newton relative and absolute tolerances
        ndivtol = 1e15 # Newton divergence tolerance
        Phi0, Pf0 = PHI[N:], _pf(U0[N:],U0[:N],ST[N:])
        while True:
            # determine primary variable for wetting phase in each cell
            # pnw and xw was updated after previous Newton iteration
            U,ST2 = update_state(U,ST[:N])
            # compute residual
            resid = residual(U, U0, ST, PHI, dt)
            if not np.all(np.isfinite(resid)):
                print(f"NaN or inf found in residual")
                fail = True
                break
            # compute residual norm
            res = norm(np.asarray(resid))
            if it == 1:
                res0 = res
            print(f'it {it} res {res:.3g} res0 {res0:.3g} res/res0 {res/res0:.3g}', end='\r')
            if it > maxit or res > ndivtol:
                fail = True
                break
            if res/res0 < nrtol or res < natol:
                converged = True
                ST[:N] = ST2
                break
            # compute jacobian matrix using sparsejac package
            A = jacfwd(residual, sparsity=spars, argnums=0)(U,U0,ST,PHI,dt)
            # convert to scipy.sparse matrix in order to use scipy.sparse solver
            cA = ssp.csc_matrix( (A.data, (A.indices[:,0], A.indices[:,1])) )
            # ilut preconditioner with default parameters
            M = ssp.linalg.LinearOperator((2*N,2*N), ssp.linalg.spilu(cA).solve)
            # solve linear system
            dU, info = ssp.linalg.bicgstab(A=cA, b=-resid, rtol=lrtol, atol=latol, M=M, maxiter=100, callback=None)
            if info != 0:
                print('\nFailed to solve linear system')
                pdb.set_trace()
            it += 1
            w = 1. # relaxation update parameter
            linesearchfail = True
            while linesearchfail and w > 1e-5:
                U1 = U + w*dU
                U1,ST2 = update_state(U1,ST[:N])
                PHI2 = _phi1(U1[N:],U1[:N],ST2, phi0=Phi0, pf0=Pf0)
                Sw = _sw1(U1[N:],U1[:N],ST2)
                if np.all(np.isfinite(U1)) and np.all(Sw > 0.0) and np.all(Sw < 1.0) and np.all(PHI2 < 1.0) and np.all(PHI2 > 0.0):
                    linesearchfail = False
                    ST[:N] = ST2
                else:
                    w *= 0.05
            if linesearchfail:
                print(f'\nLine search failed: t {t:.2g} dt {dt:.2g}')
                fail = True
                break
            else:
                U = U1
                PHI[:N] = PHI2
        if fail:
            if rst > maxrst:
                print(f'\nfail after {maxrst} restarts: t {t:.2g} dt {dt:.2g}')
                pdb.set_trace()
                exit()
            restart = True
        elif converged:
            #if np.max(PHI[:N]) > 0.95:
            #    pdb.set_trace()
            if fig is None:
                fig, axs = plt.subplots(nrows=2,ncols=4, figsize=(9, 6))
                axs[0,0].set_title('Pnw')
                axs[0,1].set_title('Sw')
                axs[0,2].set_title('Pw')
                axs[0,3].set_title('Pf')
                axs[1,0].set_title('St')
                axs[1,1].set_title('Pc')
                axs[1,2].set_title('Phi')
                for ax in axs.flatten():
                    remove_ticks(ax)
                PF, PF0 = _pf1(U[N:],U[:N],ST[:N]), _pf1(U0[N:],U0[:N],ST[:N])
                im1 = axs[0,0].imshow( U[N:].reshape((Nz,Nr)), vmax=np.max(U[N:]) )
                im2 = axs[0,1].imshow( _sw1(U[N:],U[:N],ST[:N]).reshape((Nz,Nr)), vmax=np.max(U[:N]) )
                im3 = axs[0,2].imshow( _pw1(U[N:],U[:N],ST[:N]).reshape((Nz,Nr)))
                im4 = axs[0,3].imshow( PF.reshape((Nz,Nr)), vmin=0, vmax=1 )
                im5 = axs[1,0].imshow( ST[:N].reshape((Nz,Nr)), vmin=0, vmax=1 )
                im6 = axs[1,1].imshow( Pc(_sw1(U[N:],U[:N],ST[:N])).reshape((Nz,Nr)))
                im7 = axs[1,2].imshow( _phi1(U[N:],U[:N],ST[:N],phi0=PHI[N:],pf0=PF0).reshape((Nz,Nr)))
                cbar1 = fig.colorbar(im1, ax=axs[0,0])
                cbar2 = fig.colorbar(im2, ax=axs[0,1])
                cbar3 = fig.colorbar(im3, ax=axs[0,2])
                cbar4 = fig.colorbar(im4, ax=axs[0,3])
                cbar5 = fig.colorbar(im5, ax=axs[1,0])
                cbar6 = fig.colorbar(im6, ax=axs[1,1])
                cbar7 = fig.colorbar(im7, ax=axs[1,2])
            else:
                im1.set_data(U[N:].reshape((Nz,Nr)))
                SwPlot = _sw1(U[N:],U[:N],ST[:N]).reshape((Nz,Nr))
                im2.set_data(SwPlot)
                PwPlot = _pw1(U[N:],U[:N],ST[:N]).reshape((Nz,Nr))
                im3.set_data(PwPlot)
                PfPlot = _pf1(U[N:],U[:N],ST[:N]).reshape((Nz,Nr))
                im4.set_data(PfPlot)
                PcPlot = Pc(_sw1(U[N:],U[:N],ST[:N])).reshape((Nz,Nr))
                im5.set_data(ST[:N].reshape((Nz,Nr)))
                im6.set_data(PcPlot)
                PhiPlot = _phi1(U[N:],U[:N],ST[:N],phi0=PHI[N:],pf0=_pf1(U0[N:],U0[:N],ST[:N])).reshape((Nz,Nr))
                im7.set_data(PhiPlot)
                im1.set_clim(vmax=np.max(U[N:]), vmin=np.min(U[N:]))
                im2.set_clim(vmax=np.max(SwPlot), vmin=np.min(SwPlot))
                im3.set_clim(vmax=np.max(PwPlot), vmin=np.min(PwPlot))
                im4.set_clim(vmax=np.max(PfPlot), vmin=np.min(PfPlot))
                im6.set_clim(vmax=np.max(PcPlot), vmin=np.min(PcPlot))
                im7.set_clim(vmax=np.max(PhiPlot), vmin=np.min(PhiPlot))
            plt.show(block=False)
            plt.pause(0.0001)
            U0 = np.asarray(U)
            ST[N:] = ST[:N]
            PHI[N:] = PHI[:N]
            print(f'\nt {t:.3g} dt {dt:.2g}')
            t += dt
            rst = 0
            if it < 7:
                dt = min(1.5*dt,dtmax)
            # input('to proceed press Enter')
        if restart:
            ST[:N] = ST[N:]
            PHI[:N] = PHI[N:]
            U = jnp.asarray(U0)
            dt /= 2
            rst += 1
            print(f'restart {rst} t {t:.2g} new dt {dt:.2g}')
    input('Finished, press Enter to exit')
